@@ -1,8 +1,8 @@
 #ifndef PMT_COMPILER_H
 #define PMT_COMPILER_H 1
 
+#include <libHX/option.h>
 #include "config.h"
-#include <libHX.h>
 
 #ifdef HAVE_VISIBILITY_HIDDEN
 #	define EXPORT_SYMBOL __attribute__((visibility("default")))
@@ -16,6 +16,8 @@
 #define reinterpret_cast(type, expr)    ((type)(expr))
 #define static_cast(type, expr)         ((type)(expr))
 
+struct HXbtree;
+
 /* Short wrappers */
 static inline void format_add(struct HXbtree *table, const char *key,
     const char *value)
