@@ -1,27 +1,13 @@
-/*=============================================================================
-pam_mount - misc.c
-  Copyright (C) Elvis Pfützenreuter <epx@conectiva.com>, 2000
-  Copyright © CC Computer Consultants GmbH, 2005 - 2007
-  Contact: Jan Engelhardt <jengelh [at] computergmbh de>
-  Copyright © Bastian Kleineidam <calvin [at] debian org>, 2005
-
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU Lesser General Public License as
-  published by the Free Software Foundation; either version 2.1 of
-  the License, or (at your option) any later version.
-
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-  Lesser General Public License for more details.
-
-  You should have received a copy of the GNU Lesser General Public
-  License along with this program; if not, write to:
-  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-  Boston, MA  02110-1301  USA
-
-  -- For details, see the file named "LICENSE.LGPL2"
-=============================================================================*/
+/*
+ *	Copyright (C) Elvis Pfützenreuter, 2000
+ *	Copyright © Jan Engelhardt, 2005 - 2008
+ *	Copyright © Bastian Kleineidam, 2005
+ *
+ *	This file is part of pam_mount; you can redistribute it and/or
+ *	modify it under the terms of the GNU Lesser General Public License
+ *	as published by the Free Software Foundation; either version 2.1
+ *	of the License, or (at your option) any later version.
+ */
 #include <config.h>
 #include <sys/stat.h>
 #include <assert.h>
@@ -32,6 +18,7 @@ pam_mount - misc.c
 #include <string.h>
 #include <syslog.h>
 #include <unistd.h>
+#include <libHX/list.h>
 #include <libHX.h>
 #include <pwd.h>
 #include "misc.h"
@@ -44,7 +31,7 @@ pam_mount - misc.c
 static int static_string_valid(const char *, const size_t);
 
 //-----------------------------------------------------------------------------
-/*
+/**
  * misc_dump_id - print user IDs
  */
 void misc_dump_id(const char *where)
@@ -57,7 +44,7 @@ void misc_dump_id(const char *where)
 	return;
 }
 
-/*
+/**
  * misc_log - log to syslog
  * @format:	printf(3)-style format specifier
  *
@@ -75,13 +62,13 @@ void misc_log(const char *format, ...)
 	va_copy(arg2, args);
 	if (Debug)
 		vfprintf(stderr, format, args);
-	vsyslog(LOG_AUTHPRIV | LOG_ERR, format, arg2);
+	vsyslog(LOG_AUTH | LOG_ERR, format, arg2);
 	va_end(args);
 	va_end(arg2);
 	return;
 }
 
-/*
+/**
  * misc_warn - debug logger
  * @format:	printf(3)-style format specifier
  *
@@ -99,13 +86,13 @@ void misc_warn(const char *format, ...)
 	va_start(args, format);
 	va_copy(arg2, args);
 	vfprintf(stderr, format, args);
-	vsyslog(LOG_AUTHPRIV | LOG_ERR, format, arg2);
+	vsyslog(LOG_AUTH | LOG_ERR, format, arg2);
 	va_end(args);
 	va_end(arg2);
 	return;
 }
 
-/*
+/**
  * exists -
  * @file:	file to check
  *
@@ -121,7 +108,8 @@ int exists(const char *file)
 }
 
 
-/* owns -
+/**
+ * owns -
  * @user:	user to check for
  * @file:	file to check
  *
@@ -152,7 +140,7 @@ int owns(const char *user, const char *file)
 	       !S_ISLNK(filestat.st_mode);
 }
 
-/*
+/**
  * str_to_long -
  * @n:	string to analyze
  *
@@ -177,7 +165,7 @@ long str_to_long(const char *n)
 	return val;
 }
 
-/*
+/**
  * static_string_valid -
  * @s:		string to analyze
  * @len:	maximum length of string
@@ -196,7 +184,7 @@ static int static_string_valid(const char *s, const size_t len)
 	return 0;
 }
 
-/*
+/**
  * vol_valid -
  * @v:	volume to check
  *
@@ -232,7 +220,7 @@ int vol_valid(const struct vol *v) {
 	return 1;
 }
 
-/*
+/**
  * config_valid -
  * @c:	config to check
  *
@@ -240,7 +228,8 @@ int vol_valid(const struct vol *v) {
  */
 int config_valid(const struct config *c)
 {
-	int i;
+	struct vol *vol;
+
 	if (c == NULL || c->user == NULL)
 		return 0;
 	/* bool debug */
@@ -253,13 +242,14 @@ int config_valid(const struct config *c)
 	/* optlist_t *options_require; */
 	/* optlist_t *options_allow; */
 	/* optlist_t *options_deny; */
-	for (i = 0; i < c->volcount; ++i)
-		if (!vol_valid(c->volume))
+
+	HXlist_for_each_entry(vol, &c->volume_list, list)
+		if (!vol_valid(vol))
 			return 0;
 	return 1;
 }
 
-/*
+/**
  * log_argv -
  * @argv:	argument vector
  *
@@ -287,7 +277,8 @@ void log_argv(const char *const *argv)
 	return;
 }
 
-/* add_to_argv -
+/**
+ * add_to_argv -
  * @argv:	argument vector to add to
  * @argc:	pointer to current argument count
  * @arg:	argument to add
@@ -315,21 +306,19 @@ void add_to_argv(const char **argv, int *const argc, const char *const arg,
 		l0g("too many arguments to mount command\n");
 		return;
 	}
-	if (HXformat_aprintf(vinfo, &filled, arg) == 0) {
+	if (HXformat_aprintf(vinfo, &filled, arg) == 0)
 		/*
 		 * This case may happen with e.g. %(before="-o" OPTIONS) where
 		 * OPTIONS is empty. And empty options is certainly valid.
 		 */
-		w4rn("could not fill %s\n", arg);
 		return;
-	}
 
 	argv[*argc]   = filled;
 	argv[++*argc] = NULL;
 	return;
 }
 
-/*
+/**
  * set_myuid -
  * @user:	switch to specified user
  *
@@ -351,7 +340,8 @@ void add_to_argv(const char **argv, int *const argc, const char *const arg,
 void set_myuid(const char *user)
 {
 	setsid();
-	chdir("/");
+	if (chdir("/") < 0)
+		;
 	if (user == NULL) {
 		misc_dump_id("set_myuid<pre>");
 		if (setuid(0) < 0) {
@@ -383,7 +373,7 @@ void set_myuid(const char *user)
 	return;
 }
 
-/*
+/**
  * relookup_user -
  * @user:	The user to retrieve
  *
@@ -400,7 +390,7 @@ char *relookup_user(const char *user)
 		return xstrdup(pe->pw_name);
 }
 
-/*
+/**
  * misc_add_ntdom -
  * @v:		substitution data
  * @user:	username to add
@@ -419,7 +409,8 @@ void misc_add_ntdom(struct HXbtree *v, const char *user)
 	*domain = '\0';
 
 	if ((ptr = strchr(user, '\\')) != NULL) {
-		snprintf(domain, sizeof(domain), "%.*s", ptr - user - 1, user);
+		snprintf(domain, sizeof(domain), "%.*s",
+		         static_cast(int, ptr - user - 1), user);
 		domain_user = ptr + 1;
 	} else {
 		domain_user = user;
