@@ -1,27 +1,13 @@
-/*=============================================================================
-pam_mount - mount.c
-  Copyright (C) Elvis Pfützenreuter <epx@conectiva.com>, 2000
-  Copyright © CC Computer Consultants GmbH, 2005 - 2007
-  Contact: Jan Engelhardt <jengelh [at] computergmbh de>
-  Copyright © Bastian Kleineidam <calvin [at] debian org>, 2005
-
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU Lesser General Public License as
-  published by the Free Software Foundation; either version 2.1 of
-  the License, or (at your option) any later version.
-
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-  Lesser General Public License for more details.
-
-  You should have received a copy of the GNU Lesser General Public
-  License along with this program; if not, write to:
-  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-  Boston, MA  02110-1301  USA
-
-  -- For details, see the file named "LICENSE.LGPL2"
-=============================================================================*/
+/*
+ *	Copyright © Elvis Pfützenreuter, 2000
+ *	Copyright © Jan Engelhardt, 2006 - 2008
+ *	Copyright © Bastian Kleineidam, 2005
+ *
+ *	This file is part of pam_mount; you can redistribute it and/or
+ *	modify it under the terms of the GNU Lesser General Public License
+ *	as published by the Free Software Foundation; either version 2.1
+ *	of the License, or (at your option) any later version.
+ */
 #include <config.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -46,6 +32,7 @@ pam_mount - mount.c
 #include "private.h"
 #include "readconfig.h"
 #include "spawn.h"
+#include "xstdlib.h"
 #if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
 #	include <fstab.h>
 #elif defined(__linux__)
@@ -60,12 +47,9 @@ pam_mount - mount.c
 #endif
 
 /* Functions */
-static int already_mounted(const struct config * const, const unsigned int, struct HXbtree *);
-static int check_filesystem(const struct config *, const unsigned int, struct HXbtree *, const unsigned char *, size_t);
-static int do_losetup(const struct config *, const unsigned int, struct HXbtree *, const unsigned char *, size_t);
+static int already_mounted(const struct config * const, const struct vol *, struct HXbtree *);
 static int do_unlosetup(const struct config *, struct HXbtree *);
 static int fstype_nodev(const char *);
-static void log_pm_input(const struct config * const, const unsigned int);
 static inline bool mkmountpoint(struct vol *, const char *);
 static int pipewrite(int, const void *, size_t);
 static void run_lsof(const struct config * const, struct HXbtree *);
@@ -79,14 +63,16 @@ static int split_bsd_mount(char *, const char **, const char **, const char **);
 #endif
 
 //-----------------------------------------------------------------------------
-/*
+/**
  * log_output
  * @fd:		file descriptor to read from
  * @cmsg:	conditional message
  *
  * Reads all data from @fd and logs it using w4rn(). @fd is usually connected
  * to a pipe to another process's stdout or stderr. Only if @fd actually has
- * output for us, @cmsg will be printed
+ * output for us, @cmsg will be printed.
+ *
+ * @fd will be closed.
  */
 static void log_output(int fd, const char *cmsg)
 {
@@ -99,18 +85,20 @@ static void log_output(int fd, const char *cmsg)
 	}
 
 	setvbuf(fp, NULL, _IOLBF, 0);
-	if (fgets(buf, sizeof(buf), fp) != NULL)
+	if (fgets(buf, sizeof(buf), fp) != NULL) {
 		if (cmsg != NULL)
 			w4rn("%s", cmsg);
 
-	do {
-		w4rn("%s", buf);
-	} while (fgets(buf, sizeof(buf), fp) != NULL);
+		do {
+			w4rn("%s", buf);
+		} while (fgets(buf, sizeof(buf), fp) != NULL);
+	}
+	fclose(fp);
 	return;
 }
 
-/*
- * run_lsof
+/**
+ * run_lsof -
  * @config:	current configuration
  * @vinfo:
  *
@@ -122,9 +110,13 @@ static void run_lsof(const struct config *const config,
 {
 	int _argc = 0, cstdout = -1;
 	const char *_argv[MAX_PAR + 1];
+	struct stat sb;
 	unsigned int i;
 	pid_t pid;
 
+	if (stat(HXbtree_get(vinfo, "MNTPT"), &sb) < 0 && errno == EEXIST)
+		return;
+
 	if (config->command[CMD_LSOF][0] == NULL)
 		l0g("lsof not defined in pam_mount.conf.xml\n");
 	/* FIXME: NEW */
@@ -140,31 +132,29 @@ static void run_lsof(const struct config *const config,
 	if (waitpid(pid, NULL, 0) < 0)
 		l0g("error waiting for child: %s\n", strerror(errno));
 	spawn_restore_sigchld();
-	close(cstdout);
 	return;
 }
 
 
-/* already_mounted
+/**
+ * already_mounted -
  * @config:	current config
- * @vol:	volume index into @config->volume[]
+ * @vpt:	volume descriptor
  * @vinfo:
  *
  * Checks if @config->volume[@vol] is already mounted, and returns 1 if this
  * the case, 0 if not and -1 on error.
  */
 static int already_mounted(const struct config *const config,
-    const unsigned int vol, struct HXbtree *vinfo)
+    const struct vol *vpt, struct HXbtree *vinfo)
 #if defined(__linux__)
 {
 	char dev[PATH_MAX+1] = {}, real_mpt[PATH_MAX+1];
 	struct mntent *mtab_record;
 	bool mounted = false;
 	FILE *mtab;
-	struct vol *vpt;
 
 	assert(config_valid(config));
-	vpt = &config->volume[vol];
 	vol_to_dev(dev, sizeof(dev), vpt);
 
 	if ((mtab = setmntent("/etc/mtab", "r")) == NULL) {
@@ -292,8 +282,8 @@ static int already_mounted(const struct config *const config,
 }
 #endif
 
-/*
- * vol_to_dev
+/**
+ * vol_to_dev -
  * @match:
  * @s:
  * @vol:	volume to analyze
@@ -308,8 +298,7 @@ static void vol_to_dev(char *match, size_t s, const struct vol *vol)
 
 	case CMD_NCPMOUNT:
 		snprintf(match, s, "%s/%s", vol->server,
-		         static_cast(const char *,
-		         HXbtree_get(vol->options, "user")));
+		         kvplist_get(&vol->options, "user"));
 		break;
 
 	case CMD_NFSMOUNT:
@@ -371,41 +360,12 @@ static int split_bsd_mount(char *wp, const char **fsname, const char **fspt,
 }
 #endif
 
-/*
- * optlist_to_str -
- * @optlist:	option list
- *
- * Transform the option list into a flat string. Allocates and returns the
- * string. Caller has to free it. Used for debugging.
- */
-static hmc_t *optlist_to_str(const struct HXbtree *optlist)
-{
-	const struct HXbtree_node *option;
-	hmc_t *ret = hmc_sinit("");
-	void *trav;
-
-	if (optlist == NULL)
-		return ret;
-
-	trav = HXbtrav_init(optlist);
-	while ((option = HXbtraverse(trav)) != NULL) {
-		hmc_strcat(&ret, option->key);
-		if (option->data != NULL &&
-		    *static_cast(const char *, option->data) != '\0') {
-			hmc_strcat(&ret, "=");
-			hmc_strcat(&ret, option->data);
-		}
-		hmc_strcat(&ret, ",");
-	}
-	return ret;
-}
 static void log_pm_input(const struct config *const config,
-    const unsigned int vol)
+    const struct vol *vpt)
 {
-	const struct vol *vpt = &config->volume[vol];
 	hmc_t *options;
 
-	options = optlist_to_str(vpt->options);
+	options = kvplist_to_str(&vpt->options);
 	w4rn("information for mount:\n");
 	w4rn("----------------------\n");
 	w4rn("(defined by %s)\n", vpt->globalconf ? "globalconf" : "luserconf");
@@ -422,7 +382,7 @@ static void log_pm_input(const struct config *const config,
 	return;
 }
 
-/*
+/**
  * mkmountpoint_real - create mountpoint directory
  * @volume:	volume description
  * @d:		directory
@@ -473,7 +433,7 @@ static bool mkmountpoint_real(struct vol *const volume, const char *const d)
 	return ret;
 }
 
-/*
+/**
  * mkmountpoint_pick - create mountpoint for volume
  * @volume:	volume structure
  * @d:		directory to create
@@ -507,7 +467,7 @@ static bool mkmountpoint_pick(struct vol *volume, const char *d)
 	return ret;
 }
 
-/*
+/**
  * mkmountpoint -
  *
  * Wrapper for mkmountpoint_pick(). Switch back to root user after
@@ -521,22 +481,21 @@ static inline bool mkmountpoint(struct vol *volume, const char *d)
 	return r;
 }
 
-/*
- * do_unmount
+/**
+ * do_unmount -
  * @config:	current config
- * @vol:	volume index into @config->vol[]
+ * @vpt:	volume descriptor
  * @vinfo:
  * @password:	always %NULL
  *
  * Returns zero on error, positive non-zero for success.
  */
-int do_unmount(const struct config *config, const unsigned int vol,
+int do_unmount(const struct config *config, struct vol *vpt,
     struct HXbtree *vinfo, const char *const password)
 {
 	int child_exit, _argc = 0, ret = 1, cstderr = -1;
 	pid_t pid = -1;
 	const char *_argv[MAX_PAR + 1];
-	const struct vol *vpt;
 	unsigned int i;
 	int type;
 
@@ -544,8 +503,6 @@ int do_unmount(const struct config *config, const unsigned int vol,
 	assert(vinfo != NULL);
 	assert(password == NULL);	/* password should point to NULL for unmounting */
 
-	vpt = &config->volume[vol];
-
 	if (Debug)
 		/*
 		 * Often, a process still exists with ~ as its pwd after
@@ -592,7 +549,6 @@ int do_unmount(const struct config *config, const unsigned int vol,
 		goto out;
 	}
 	log_output(cstderr, "umount errors:\n");
-	close(cstderr);
 	w4rn("waiting for umount\n");
 	if (waitpid(pid, &child_exit, 0) < 0) {
 		l0g("error waiting for child: %s\n", strerror(errno));
@@ -641,7 +597,7 @@ static int pipewrite(int fd, const void *buf, size_t count)
 	return fnval;
 }
 
-static int do_losetup(const struct config *config, const unsigned int vol,
+static int do_losetup(const struct config *config, const struct vol *vpt,
     struct HXbtree *vinfo, const unsigned char *password, size_t password_len)
 {
 /* PRE:    config points to a valid struct config
@@ -654,7 +610,6 @@ static int do_losetup(const struct config *config, const unsigned int vol,
 	int ret = 1, child_exit, _argc = 0, cstdin = -1, cstderr = -1;
 	const char *_argv[MAX_PAR + 1];
 	const char *cipher, *keybits;
-	const struct vol *vpt;
 	unsigned int i;
 
 	assert(config_valid(config));
@@ -663,9 +618,8 @@ static int do_losetup(const struct config *config, const unsigned int vol,
 	/* password_len is unsigned */
 	assert(password_len <= MAX_PAR + EVP_MAX_BLOCK_LENGTH);
 
-	vpt     = &config->volume[vol];
-	cipher  = HXbtree_get(vpt->options, "encryption");
-	keybits = HXbtree_get(vpt->options, "keybits");
+	cipher  = kvplist_get(&vpt->options, "encryption");
+	keybits = kvplist_get(&vpt->options, "keybits");
 
 	if (config->command[CMD_LOSETUP][0] == NULL) {
 		l0g("losetup not defined in pam_mount.conf.xml\n");
@@ -693,7 +647,6 @@ static int do_losetup(const struct config *config, const unsigned int vol,
 	}
 	close(cstdin);
 	log_output(cstderr, "losetup errors:\n");
-	close(cstderr);
 	w4rn("waiting for losetup\n");
 	if (waitpid(pid, &child_exit, 0) < 0) {
 		l0g("error waiting for child: %s\n", strerror(errno));
@@ -740,7 +693,7 @@ static int do_unlosetup(const struct config *config, struct HXbtree *vinfo)
 	return !WEXITSTATUS(child_exit);
 }
 
-static int check_filesystem(const struct config *config, const unsigned int vol,
+static int check_filesystem(const struct config *config, const struct vol *vpt,
     struct HXbtree *vinfo, const unsigned char *password, size_t password_len)
 {
 /* PRE:    config points to a valid struct config
@@ -753,7 +706,6 @@ static int check_filesystem(const struct config *config, const unsigned int vol,
 	int child_exit, _argc = 0, cstdout = -1, cstderr = -1;
 	const char *_argv[MAX_PAR + 1];
 	const char *fsck_target;
-	const struct vol *vpt;
 	unsigned int i;
 
 	assert(config_valid(config));
@@ -762,7 +714,6 @@ static int check_filesystem(const struct config *config, const unsigned int vol,
 	assert(password_len >= 0 &&
 	       password_len <= MAX_PAR + EVP_MAX_BLOCK_LENGTH);
 
-	vpt = &config->volume[vol];
 	fsck_target = vpt->volume;
 
 	if (config->command[CMD_FSCK][0] == NULL) {
@@ -770,17 +721,17 @@ static int check_filesystem(const struct config *config, const unsigned int vol,
 		return 0;
 	}
 
-	if (HXbtree_find(vpt->options, "bind") != NULL ||
-	    HXbtree_find(vpt->options, "move") != NULL ||
+	if (kvplist_contains(&vpt->options, "bind") ||
+	    kvplist_contains(&vpt->options, "move") ||
 	    fstype_nodev(vpt->fstype) != 0)
 		return 1;
 
-	if (HXbtree_find(vpt->options, "loop") != NULL) {
-		if (!do_losetup(config, vol, vinfo, password, password_len))
+	if (kvplist_contains(&vpt->options, "loop")) {
+		if (!do_losetup(config, vpt, vinfo, password, password_len))
 			return 0;
 		fsck_target = config->fsckloop;
 	} else {
-		hmc_t *options = optlist_to_str(vpt->options);
+		hmc_t *options = kvplist_to_str(&vpt->options);
 		w4rn("volume not a loopback (options: %s)\n", options);
 		hmc_free(options);
 	}
@@ -797,12 +748,11 @@ static int check_filesystem(const struct config *config, const unsigned int vol,
 	/* stdout and stderr must be logged for fsck */
 	log_output(cstdout, NULL);
 	log_output(cstderr, NULL);
-	close(cstderr);
 	w4rn("waiting for filesystem check\n");
 	if (waitpid(pid, &child_exit, 0) < 0)
 		l0g("error waiting for child: %s\n", strerror(errno));
 	spawn_restore_sigchld();
-	if (HXbtree_find(vpt->options, "loop") != NULL)
+	if (kvplist_contains(&vpt->options, "loop"))
 		if (!do_unlosetup(config, vinfo))
 			return 0;
 	/*
@@ -816,16 +766,16 @@ static int check_filesystem(const struct config *config, const unsigned int vol,
 #endif
 }
 
-/*
+/**
  * do_mount -
  * @config:	current config
- * @vol:	volume index into @config->vol[]
+ * @vpt:	volume descriptor
  * @vinfo:
  * @password:
  *
  * Returns zero on error, positive non-zero for success.
  */
-int do_mount(const struct config *config, const unsigned int vol,
+int do_mount(const struct config *config, struct vol *vpt,
     struct HXbtree *vinfo, const char *password)
 {
 	const char *_argv[MAX_PAR + 1];
@@ -834,7 +784,6 @@ int do_mount(const struct config *config, const unsigned int vol,
 	int _argc = 0, child_exit = 0, cstdin = -1, cstderr = -1;
 	char *mount_user;
 	pid_t pid = -1;
-	struct vol *vpt;
 	unsigned int i;
 	int ret;
 
@@ -842,8 +791,7 @@ int do_mount(const struct config *config, const unsigned int vol,
 	assert(vinfo != NULL);
 	assert(password != NULL);
 
-	vpt = &config->volume[vol];
-	ret = already_mounted(config, vol, vinfo);
+	ret = already_mounted(config, vpt, vinfo);
 	if (ret == -1) {
 		l0g("could not determine if %s is already mounted, "
 		    "failing\n", vpt->volume);
@@ -893,7 +841,7 @@ int do_mount(const struct config *config, const unsigned int vol,
 		 * standard string MAX_PAR + 1 in this case
 		 */
 		strncpy(signed_cast(char *, _password), password, MAX_PAR);
-		_password[MAX_PAR] = '\0';
+		_password[MAX_PAR-1] = '\0';
 		_password_len = strlen(password);
 	}
 	w4rn("about to start building mount command\n");
@@ -902,12 +850,18 @@ int do_mount(const struct config *config, const unsigned int vol,
 	   l0g("volume type (%d) is unknown\n", vpt->type);
 	   return 0;
 	 */
+
+	if (vpt->uses_ssh)
+		for (i = 0; config->command[CMD_FD0SSH][i] != NULL; ++i)
+			add_to_argv(_argv, &_argc,
+			            config->command[CMD_FD0SSH][i], vinfo);
+
 	for (i = 0; config->command[vpt->type][i] != NULL; ++i)
 		add_to_argv(_argv, &_argc,
 		            config->command[vpt->type][i], vinfo);
 
 	if (vpt->type == CMD_LCLMOUNT &&
-	    !check_filesystem(config, vol, vinfo, _password, _password_len))
+	    !check_filesystem(config, vpt, vinfo, _password, _password_len))
 		l0g("error checking filesystem but will continue\n");
 	/* send password down pipe to mount process */
 	if (vpt->type == CMD_SMBMOUNT || vpt->type == CMD_CIFSMOUNT)
@@ -929,7 +883,6 @@ int do_mount(const struct config *config, const unsigned int vol,
 	/* Paranoia? */
 	memset(_password, 0, sizeof(_password));
 	log_output(cstderr, "mount errors:\n");
-	close(cstderr);
 	w4rn("waiting for mount\n");
 	if (waitpid(pid, &child_exit, 0) < 0) {
 		spawn_restore_sigchld();
@@ -939,36 +892,35 @@ int do_mount(const struct config *config, const unsigned int vol,
 
 	spawn_restore_sigchld();
 	if (Debug)
-		system("df -Ta");
+		if (system("df -Ta") < 0)
+			;
 
 	/* pass on through the result from the umount process */
 	return !WEXITSTATUS(child_exit);
 }
 
-/*
+/**
  * mount_op -
  * @mnt:	function to execute mount operations (do_mount or do_unmount)
- * @config:
- * @vol:	volume index into @config->volume[]
+ * @config:	current configuration
+ * @vpt:	volume descriptor
  * @password:	password string (may be %NULL on unmount)
  *
  * Returns zero on error, positive non-zero for success.
  * Note: Checked by volume_record_sane() and read_volume()
  */
 int mount_op(mount_op_fn_t *mnt, const struct config *config,
-    const unsigned int vol, const char *password)
+    struct vol *vpt, const char *password)
 {
 	int fnval;
 	struct HXbtree *vinfo;
-	const struct vol *vpt;
 	struct passwd *pe;
 	hmc_t *options;
 
 	assert(config_valid(config));
 
-	vpt = &config->volume[vol];
-
-	vinfo = HXformat_init();
+	if ((vinfo = HXformat_init()) == NULL)
+		return 0;
 	format_add(vinfo, "MNTPT",    vpt->mountpoint);
 	format_add(vinfo, "FSCKLOOP", config->fsckloop);
 	format_add(vinfo, "FSTYPE",   vpt->fstype);
@@ -981,24 +933,25 @@ int mount_op(mount_op_fn_t *mnt, const struct config *config,
 		w4rn("getpwnam(\"%s\") failed: %s\n",
 		     Config.user, strerror(errno));
 	} else {
-		HXformat_add(vinfo, "USERUID", &pe->pw_uid, HXTYPE_LONG);
-		HXformat_add(vinfo, "USERGID", &pe->pw_gid, HXTYPE_LONG);
+		unsigned int uid = pe->pw_uid, gid = pe->pw_gid;
+		HXformat_add(vinfo, "USERUID", &uid, HXTYPE_UINT);
+		HXformat_add(vinfo, "USERGID", &gid, HXTYPE_UINT);
 	}
 
 	/* FIXME: should others remain undefined if == ""? */
-	options = optlist_to_str(vpt->options);
+	options = kvplist_to_str(&vpt->options);
 	format_add(vinfo, "OPTIONS", options);
 
 	if (Debug)
-		log_pm_input(config, vol);
+		log_pm_input(config, vpt);
 
-	fnval = (*mnt)(config, vol, vinfo, password);
+	fnval = (*mnt)(config, vpt, vinfo, password);
 	hmc_free(options);
 	HXformat_free(vinfo);
 	return fnval;
 }
 
-/*
+/**
  * fstype_nodev -
  * @name:	fstype to check
  *
@@ -1028,7 +981,7 @@ static int fstype_nodev(const char *name) {
 }
 
 #ifdef HAVE_STRUCT_LOOP_INFO64_LO_FILE_NAME
-/*
+/**
  * loop_bk -
  * @filename:	block device to query
  * @i:		pointer to result storage
@@ -1051,5 +1004,3 @@ static inline const char *loop_bk(const char *filename, struct loop_info64 *i)
 	return signed_cast(char *, i->lo_file_name);
 }
 #endif
-
-//=============================================================================
