@@ -1,27 +1,14 @@
-/*=============================================================================
-pam_mount - pam_mount.c
-  Copyright (C) Elvis Pfützenreuter <epx@conectiva.com>, 2000
-  Copyright © CC Computer Consultants GmbH, 2005 - 2007
-  Contact: Jan Engelhardt <jengelh [at] computergmbh de>
-  Copyright © Bastian Kleineidam <calvin [at] debian org>, 2005
-
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU Lesser General Public License as
-  published by the Free Software Foundation; either version 2.1 of
-  the License, or (at your option) any later version.
-
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-  Lesser General Public License for more details.
-
-  You should have received a copy of the GNU Lesser General Public
-  License along with this program; if not, write to:
-  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-  Boston, MA  02110-1301  USA
-
-  -- For details, see the file named "LICENSE.LGPL2"
-=============================================================================*/
+/*
+ *	pam_mount
+ *	Copyright (C) Elvis Pfützenreuter <epx@conectiva.com>, 2000
+ *	Copyright © Jan Engelhardt, 2005 - 2008
+ *	Copyright © Bastian Kleineidam, 2005
+ *
+ *	This file is part of pam_mount; you can redistribute it and/or
+ *	modify it under the terms of the GNU Lesser General Public License
+ *	as published by the Free Software Foundation; either version 2.1
+ *	of the License, or (at your option) any later version.
+ */
 #define PAM_SM_SESSION 1
 #define PAM_SM_PASSWORD 1
 
@@ -47,6 +34,10 @@ pam_mount - pam_mount.c
 #include "spawn.h"
 #include "xstdlib.h"
 
+#ifndef PAM_EXTERN
+#	define PAM_EXTERN
+#endif
+
 #if defined(__OpenBSD__) || defined(__NetBSD__) || defined(__APPLE__)
 #	define CONFIGFILE "/etc/pam_mount.conf.xml"
 #else
@@ -64,12 +55,12 @@ static int read_password(pam_handle_t *, const char *, char **);
 
 /* Variables */
 static const char *envpath_saved;
-bool Debug = true;
+unsigned int Debug = true;
 struct config Config = {};
 struct pam_args Args = {};
 
 //-----------------------------------------------------------------------------
-/*
+/**
  * parse_pam_args -
  * @argv:	NULL-terminated argument vector
  * @argc:	number of elements in @argc
@@ -96,13 +87,15 @@ static void parse_pam_args(int argc, const char **argv)
 			Args.auth_type = SOFT_TRY_PASS;
 		else if (strcmp("nullok", argv[i]) == 0)
 			Args.nullok = true;
+		else if (strcmp("debug", argv[i]) == 0)
+			Debug = true;
 		else
-			w4rn("bad pam_mount option\n");
+			w4rn("bad pam_mount option \"%s\"\n", argv[i]);
 	}
 	return;
 }
 
-/*
+/**
  * clean_config -
  * @pamh:	PAM handle
  * @data:	custom data pointer
@@ -118,7 +111,7 @@ static void clean_config(pam_handle_t *pamh, void *data, int err)
 	return;
 }
 
-/*
+/**
  * clean_system_authtok -
  * @pamh:	PAM handle
  * @data:	custom data pointer
@@ -141,7 +134,7 @@ static void clean_system_authtok(pam_handle_t *pamh, void *data, int errcode)
 	return;
 }
 
-/*
+/**
  * converse -
  * @pamh:	PAM handle
  * @nargs:	number of messages
@@ -180,7 +173,7 @@ static int converse(pam_handle_t *pamh, int nargs,
 	return retval; /* propagate error status */
 }
 
-/*
+/**
  * read_password -
  * @pamh:	PAM handle
  * @prompt:	a prompt message
@@ -214,6 +207,7 @@ static int read_password(pam_handle_t *pamh, const char *prompt, char **pass)
 static int common_init(pam_handle_t *pamh, int argc, const char **argv)
 {
 	const char *pam_user;
+	char buf[8];
 	int ret;
 
 	initconfig(&Config);
@@ -244,10 +238,12 @@ static int common_init(pam_handle_t *pamh, int argc, const char **argv)
 	if (!readconfig(CONFIGFILE, true, &Config))
 		return PAM_SERVICE_ERR;
 
+	snprintf(buf, sizeof(buf), "%u", Debug);
+	setenv("_PMT_DEBUG_LEVEL", buf, true);
 	return -1;
 }
 
-/*
+/**
  * pam_sm_authenticate -
  * @pamh:	PAM handle
  * @flags:	PAM flags
@@ -323,7 +319,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_authenticate(pam_handle_t *pamh, int flags,
 	return ret;
 }
 
-/*
+/**
  * On login, $PATH is correctly set to ENV_ROOTPATH (from /etc/login.defs),
  * while on logout, it happens to be ENV_PATH only. This is problematic,
  * since some programs are in /sbin and /usr/sbin which is
@@ -347,7 +343,7 @@ static void envpath_restore(void)
 	return;
 }
 
-/*
+/**
  * modify_pm_count -
  * @config:
  * @user:
@@ -382,7 +378,10 @@ static int modify_pm_count(struct config *config, char *user,
 		fnval = -1;
 		goto nosigactout;
 	}
-	vinfo = HXformat_init();
+	if ((vinfo = HXformat_init()) == NULL) {
+		fnval = -1;
+		goto out;
+	}
 	format_add(vinfo, "USER", user);
 	format_add(vinfo, "OPERATION", operation);
 	misc_add_ntdom(vinfo, user);
@@ -431,7 +430,7 @@ static int modify_pm_count(struct config *config, char *user,
 	return fnval;
 }
 
-/*
+/**
  * pam_sm_open_session -
  * @pamh:	PAM handle
  * @flags:	PAM flags
@@ -445,7 +444,7 @@ static int modify_pm_count(struct config *config, char *user,
 PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
     int argc, const char **argv)
 {
-	unsigned int vol;
+	struct vol *vol;
 	int ret;
 	unsigned int krb5_set;
 	char *system_authtok;
@@ -491,7 +490,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 	} else
 		w4rn("%s does not exist or is not owned by user\n",
 		     Config.luserconf);
-	if (Config.volcount <= 0) {
+	if (Config.volume_list.items == 0) {
 		w4rn("no volumes to mount\n");
 		ret = PAM_SUCCESS;
 		goto out;
@@ -520,7 +519,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 	misc_dump_id("Session open");
 
 	envpath_init(Config.path);
-	for (vol = 0; vol < Config.volcount; ++vol) {
+	HXlist_for_each_entry(vol, &Config.volume_list, list) {
 		/*
 		 * luserconf_volume_record_sane() is called here so that a user
 		 * can nest loopback images. otherwise ownership tests will
@@ -529,13 +528,13 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 		 */
 		if (!volume_record_sane(&Config, vol))
 			continue;
-		if (!Config.volume[vol].globalconf &&
+		if (!vol->globalconf &&
 		    !luserconf_volume_record_sane(&Config, vol))
 			continue;
 		w4rn("about to perform mount operations\n");
 
 		if (!mount_op(do_mount, &Config, vol, system_authtok)) {
-			l0g("mount of %s failed\n", Config.volume[vol].volume);
+			l0g("mount of %s failed\n", vol->volume);
 			ret = PAM_SERVICE_ERR;
 		}
 	}
@@ -549,7 +548,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 	return ret;
 }
 
-/*
+/**
  * pam_sm_chauthtok -
  * @pamh:	PAM handle
  * @flags:	PAM flags
@@ -564,7 +563,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_chauthtok(pam_handle_t *pamh, int flags,
 	return pam_sm_authenticate(pamh, flags, argc, argv);
 }
 
-/*
+/**
  * pam_sm_close_session -
  * @pamh:	PAM handle
  * @flags:	PAM flags
@@ -581,13 +580,14 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_chauthtok(pam_handle_t *pamh, int flags,
 PAM_EXTERN EXPORT_SYMBOL int pam_sm_close_session(pam_handle_t *pamh,
     int flags, int argc, const char **argv)
 {
-	int ret = PAM_SUCCESS, vol;
+	int ret = PAM_SUCCESS;
+	struct vol *vol;
 	const char *pam_user = NULL;
 
 	assert(pamh != NULL);
 
 	w4rn("received order to close things\n");
-	if (Config.volcount == 0) {
+	if (Config.volume_list.items == 0) {
 		w4rn("No volumes to umount\n");
 		goto out;
 	}
@@ -617,11 +617,11 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_close_session(pam_handle_t *pamh,
 
 	envpath_init(Config.path);
 	if (modify_pm_count(&Config, Config.user, "-1") <= 0) {
-		for (vol = Config.volcount - 1; vol >= 0; --vol) {
+		HXlist_for_each_entry_rev(vol, &Config.volume_list, list) {
 			w4rn("going to unmount\n");
 			if (!mount_op(do_unmount, &Config, vol, NULL))
 				l0g("unmount of %s failed\n",
-				    Config.volume[vol].volume);
+				    vol->volume);
 		}
 	} else {
 		w4rn("%s seems to have other remaining open sessions\n",
@@ -636,7 +636,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_close_session(pam_handle_t *pamh,
 	return ret;
 }
 
-/*
+/**
  * pam_sm_setcred -
  * @pamh:	PAM handle
  * @flags:	PAM flags
@@ -651,7 +651,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_setcred(pam_handle_t *pamh, int flags,
 	return PAM_SUCCESS;
 }
 
-/*
+/**
  * pam_sm_acct_mgmt -
  * @pamh:	PAM handle
  * @flags:	PAM flags
@@ -680,5 +680,3 @@ EXPORT_SYMBOL struct pam_module _pam_mount_modstruct = {
 };
 
 #endif
-
-//=============================================================================
