@@ -1,6 +1,7 @@
 #ifndef PMT_PAM_MOUNT_H
 #define PMT_PAM_MOUNT_H 1
 
+#include <libHX/list.h>
 #include <stdbool.h>
 
 struct config;
@@ -17,6 +18,11 @@ struct pam_args {
 	bool nullok;
 };
 
+struct kvp {
+	char *key, *value;
+	struct HXlist_head list;
+};
+
 /*
  *	PAM_MOUNT.C
  */
@@ -24,7 +30,7 @@ struct pam_args {
 	/* Workaround Xserver issue */
 #	define Debug pmt_debug
 #endif
-extern bool Debug;
+extern unsigned int Debug;
 extern struct config Config;
 extern struct pam_args Args;
 
