@@ -1,27 +1,14 @@
-/*=============================================================================
-pam_mount - pmvarrun.c
-  Copyright © W. Michael Petullo <mike@flyn.org>, 2004
-  Copyright © CC Computer Consultants GmbH, 2005 - 2007
-  Contact: Jan Engelhardt <jengelh [at] computergmbh de>
-  Copyright © Bastian Kleineidam <calvin [at] debian org>, 2005
-
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU Lesser General Public License as
-  published by the Free Software Foundation; either version 2.1 of
-  the License, or (at your option) any later version.
-
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-  Lesser General Public License for more details.
-
-  You should have received a copy of the GNU Lesser General Public
-  License along with this program; if not, write to:
-  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-  Boston, MA  02110-1301  USA
-
-  -- For details, see the file named "LICENSE.LGPL2"
-=============================================================================*/
+/*
+ *	pam_mount
+ *	Copyright © W. Michael Petullo <mike@flyn.org>, 2004
+ *	Copyright © Jan Engelhardt, 2005 - 2008
+ *	Copyright © Bastian Kleineidam <calvin [at] debian org>, 2005
+ *
+ *	This file is part of pam_mount; you can redistribute it and/or
+ *	modify it under the terms of the GNU Lesser General Public License
+ *	as published by the Free Software Foundation; either version 2.1
+ *	of the License, or (at your option) any later version.
+ */
 /*
 pmvarrun.c -- Updates /var/run/pam_mount/<user>.
     A seperate program is needed so that /var/run/pam_mount/<user> may be
@@ -41,6 +28,7 @@ pmvarrun.c -- Updates /var/run/pam_mount/<user>.
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <libHX/string.h>
 #include <pwd.h>
 #include "compiler.h"
 #include "misc.h"
@@ -70,10 +58,10 @@ static void usage(int, const char *);
 static int write_count(int, long, const char *);
 
 /* Variables */
-bool Debug = false;
+unsigned int Debug = false;
 
 //-----------------------------------------------------------------------------
-/*
+/**
  * usage - display help
  * @exitcode:	numeric value we will be exiting with
  * @error:	descriptive error string
@@ -88,7 +76,7 @@ static void usage(const int exitcode, const char *error)
 	exit(exitcode);
 }
 
-/*
+/**
  * set_defaults -
  * @settings:	pointer to settings structure
  */
@@ -100,8 +88,7 @@ static void set_defaults(struct settings *settings)
 }
 
 /*
- * from https://vitalnix.svn.sourceforge.net/svnroot/vitalnix/
- * /trunk/src/libvxutil/util.c
+ * from git://dev.medozas.de/vitalnix /src/libvxutil/util.c
  */
 static bool valid_username(const char *n)
 {
@@ -128,8 +115,8 @@ static bool valid_username(const char *n)
 	return true;
 }
 
-/*
- * parse_args
+/**
+ * parse_args -
  * @argc:	number of elements in @argv
  * @argv:	NULL-terminated argument vector
  * @settings:	pointer to settings structure
@@ -171,7 +158,7 @@ static void parse_args(int argc, const char **argv, struct settings *settings)
 	return;
 }
 
-/*
+/**
  * modify_pm_count -
  * @user:	user to poke on
  * @amount:	increment (usually -1, 0 or +1)
@@ -255,8 +242,8 @@ int main(int argc, const char **argv)
 }
 
 //-----------------------------------------------------------------------------
-/*
- * create_var_run
+/**
+ * create_var_run -
  *
  * Creates the /var/run/pam_mount directory required by pmvarrun and sets
  * proper permissions on it.
@@ -292,7 +279,7 @@ static int create_var_run(void)
 	return 1;
 }
 
-/*
+/**
  * open_and_lock -
  * @filename:	file to open
  *
@@ -368,7 +355,7 @@ static int open_and_lock(const char *filename, long uid) {
 	return fd;
 }
 
-/*
+/**
  * read_current_count -
  * @fd:	file descriptor to read from
  *
@@ -402,7 +389,7 @@ static long read_current_count(int fd, const char *filename) {
 	return ret;
 }
 
-/*
+/**
  * write_count -
  * @fd:		file descriptor to write to
  * @nv:		new value to write
@@ -415,9 +402,18 @@ static int write_count(int fd, long nv, const char *filename) {
 	char buf[ASCIIZ_LLX];
 	int wrt, len, ret;
 
-	if (nv <= 0 && unlink(filename) < 0) {
-		l0g("could not unlink %s: %s\n", filename, strerror(errno));
-		return 1; /* let user log in */
+	if (nv <= 0) {
+		if (unlink(filename) >= 0)
+			return true;
+		if (errno != EPERM)
+			l0g("could not unlink %s: %s\n", filename, strerror(errno));
+		/*
+		 * Fallback to just blanking the file. This can happen when
+		 * pmvarrun is called as unprivileged user.
+		 */
+		if (ftruncate(fd, 0) < 0)
+			w4rn("truncate failed: %s\n", strerror(errno));
+		return true;
 	}
 
 	if ((ret = lseek(fd, 0, SEEK_SET)) != 0) {
@@ -431,12 +427,14 @@ static int write_count(int fd, long nv, const char *filename) {
 		ret = -errno;
 		l0g("wrote %d of %d bytes; write error on %s: %s\n",
 		    (wrt < 0) ? 0 : wrt, len, filename, strerror(errno));
-		close(fd);
 		return ret;
 	}
 
-	ftruncate(fd, len);
+	if (ftruncate(fd, len) < 0) {
+		ret = -errno;
+		l0g("truncate failed: %s\n", strerror(errno));
+		return ret;
+	}
+
 	return 1;
 }
-
-//=============================================================================
