@@ -1,25 +1,11 @@
-/*=============================================================================
-pam_mount - rdconf1.c
-  Copyright © CC Computer Consultants GmbH, 2006 - 2007
-  Contact: Jan Engelhardt <jengelh [at] computergmbh de>
-
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU Lesser General Public License as
-  published by the Free Software Foundation; either version 2.1 of
-  the License, or (at your option) any later version.
-
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-  Lesser General Public License for more details.
-
-  You should have received a copy of the GNU Lesser General Public
-  License along with this program; if not, write to:
-  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-  Boston, MA  02110-1301  USA
-
-  -- For details, see the file named "LICENSE.LGPL2"
-=============================================================================*/
+/*
+ *	Copyright © Jan Engelhardt, 2006 - 2008
+ *
+ *	This file is part of pam_mount; you can redistribute it and/or
+ *	modify it under the terms of the GNU Lesser General Public License
+ *	as published by the Free Software Foundation; either version 2.1
+ *	of the License, or (at your option) any later version.
+ */
 #include <ctype.h>
 #include <errno.h>
 #include <grp.h>
@@ -34,7 +20,9 @@ pam_mount - rdconf1.c
 #elif defined(__linux__)
 #	include <mntent.h>
 #endif
-#include <libHX.h>
+#include <libHX/arbtree.h>
+#include <libHX/option.h>
+#include <libHX/string.h>
 #include "compiler.h"
 #include "misc.h"
 #include "pam_mount.h"
@@ -55,13 +43,6 @@ enum fstab_field {
 	FSTAB_OPTS,
 };
 
-enum wildcard_type {
-	WC_NONE,
-	WC_ANYUSER,
-	WC_PGRP,    /* as primary group */
-	WC_SGRP,    /* in secondary group */
-};
-
 enum {
 	CMDA_AUTHPW,
 	CMDA_SESSIONPW,
@@ -79,23 +60,18 @@ struct pmt_command {
 	const char *fs, *command_name, *def[MAX_PAR + 1];
 };
 
-struct volume_attrs {
-	char *user, *pgrp, *sgrp, *fstype, *server, *path, *mntpt,
-	     *options, *fskeycipher, *fskeypath;
-	unsigned int invert;
-};
-
 /* Functions */
 static char *expand_home(const char *, char *, size_t);
 static char *expand_user(const char *, char *, size_t);
 static inline int strcmp_1u(const xmlChar *, const char *);
+static int rc_volume_cond_ext(const struct passwd *, xmlNode *);
 
 /* Variables */
 static const struct callbackmap cf_tags[];
 static const struct pmt_command default_command[];
 
 //-----------------------------------------------------------------------------
-/*
+/**
  * expandconfig -
  * @config:	configuration structure
  *
@@ -105,11 +81,8 @@ bool expandconfig(const struct config *config)
 {
 	const char *u = config->user;
 	struct vol *vpt;
-	unsigned int i;
-
-	for (i = 0; i < config->volcount; ++i) {
-		vpt = &config->volume[i];
 
+	HXlist_for_each_entry(vpt, &config->volume_list, list) {
 		if (expand_home(u, vpt->mountpoint, sizeof(vpt->mountpoint)) == NULL ||
 		    expand_user(u, vpt->mountpoint, sizeof(vpt->mountpoint)) == NULL ||
 		    expand_home(u, vpt->volume, sizeof(vpt->volume)) == NULL ||
@@ -127,7 +100,7 @@ bool expandconfig(const struct config *config)
 	return true;
 }
 
-/*
+/**
  * freeconfig -
  * @config:	config struct
  *
@@ -135,6 +108,7 @@ bool expandconfig(const struct config *config)
  */
 void freeconfig(struct config *config)
 {
+	struct vol *vol, *next;
 	unsigned int i, j;
 
 	for (i = 0; i < _CMD_MAX; ++i) {
@@ -143,8 +117,10 @@ void freeconfig(struct config *config)
 			config->command[i][j] = NULL;
 	}
 
-	for (i = 0; i < config->volcount; ++i)
-		HXbtree_free(config->volume[i].options);
+	HXlist_for_each_entry_safe(vol, next, &config->volume_list, list) {
+		kvplist_genocide(&vol->options);
+		free(vol);
+	}
 
 	HXbtree_free(config->options_allow);
 	HXbtree_free(config->options_require);
@@ -160,17 +136,17 @@ void initconfig(struct config *config)
 {
 	unsigned int i, j;
 	static const unsigned int flags =
-		HXBT_MAP | HXBT_CKEY | HXBT_CDATA | HXBT_SCMP | HXBT_CID;
+		HXBT_MAP | HXBT_CKEY | HXBT_SCMP | HXBT_CID;
 
 	memset(config, 0, sizeof(*config));
 	config->debug      = true;
 	config->mkmntpoint = true;
 	strcpy(config->fsckloop, "/dev/loop7");
 
-	config->msg_authpw    = HX_strdup("pam_mount password:");
-	config->msg_sessionpw = HX_strdup("reenter password for pam_mount:");
+	config->msg_authpw    = xstrdup("pam_mount password:");
+	config->msg_sessionpw = xstrdup("reenter password for pam_mount:");
 
-	config->path = HX_strdup("/sbin:/bin:/usr/sbin:/usr/bin:"
+	config->path = xstrdup("/sbin:/bin:/usr/sbin:/usr/bin:"
 	               "/usr/local/sbin:/usr/local/bin");
 
 	for (i = 0; default_command[i].type != -1; ++i)
@@ -181,6 +157,8 @@ void initconfig(struct config *config)
 	config->options_allow   = HXbtree_init(flags);
 	config->options_require = HXbtree_init(flags);
 	config->options_deny    = HXbtree_init(flags);
+
+	HXclist_init(&config->volume_list);
 	return;
 }
 
@@ -216,7 +194,7 @@ bool readconfig(const char *file, bool global_conf, struct config *config)
 }
 
 //-----------------------------------------------------------------------------
-/*
+/**
  * expand_home -
  * @user:	username to use for home directory lookup
  * @path:	pathname to expand
@@ -250,7 +228,7 @@ static char *expand_home(const char *user, char *path, size_t size)
 	return path;
 }
 
-/*
+/**
  * expand_user -
  * @user:	username to substitue for placeholder
  * @dest:	buffer to operate on
@@ -269,7 +247,8 @@ static char *expand_user(const char *user, char *dest, size_t size)
 	if (dest == NULL)
 		l0g("expand_user_wildcard(dest=NULL), please fix\n");
 
-	vinfo = HXformat_init();
+	if ((vinfo = HXformat_init()) == NULL)
+		return NULL;
 	HXformat_add(vinfo, "USER", user, HXTYPE_STRING);
 	misc_add_ntdom(vinfo, user);
 	HXformat_sprintf(vinfo, dest, size, dest);
@@ -277,7 +256,7 @@ static char *expand_user(const char *user, char *dest, size_t size)
 	return dest;
 }
 
-/*
+/**
  * fstab_value -
  * @volume:	path to volume
  * @field:	-
@@ -372,7 +351,7 @@ static int fstab_value(const char *volume, const enum fstab_field field,
 	return 1;
 }
 
-/*
+/**
  * get_next_argument -
  * @sptr:	pointer to pointer to writable string
  *
@@ -428,16 +407,26 @@ static inline char *get_next_argument(char **sptr)
 
 static inline bool parse_bool(const char *s)
 {
-	return strcasecmp(s, "yes") == 0 || strcasecmp(s, "on") == 0||
+	return strcasecmp(s, "yes") == 0 || strcasecmp(s, "on") == 0 ||
 	       strcasecmp(s, "true") == 0 || strcmp(s, "1") == 0;
 }
 
+static inline bool parse_bool_f(char *s)
+{
+	bool ret;
+	if (s == NULL)
+		return false;
+	ret = parse_bool(s);
+	free(s);
+	return ret;
+}
+
 static inline int strcmp_1u(const xmlChar *a, const char *b)
 {
 	return strcmp(reinterpret_cast(const char *, a), b);
 }
 
-/*
+/**
  * user_in_sgrp -
  * @user:	user to check
  * @grp:	group name
@@ -446,7 +435,7 @@ static inline int strcmp_1u(const xmlChar *a, const char *b)
  * no match was found, positive non-zero on success or negative non-zero on
  * failure.
  */
-static bool user_in_sgrp(const char *user, const char *grp)
+static bool user_in_sgrp(const char *user, const char *grp, bool icase)
 {
 	struct group *gent;
 	const char **wp;
@@ -458,7 +447,8 @@ static bool user_in_sgrp(const char *user, const char *grp)
 
 	wp = const_cast(const char **, gent->gr_mem);
 	while (wp != NULL && *wp != NULL) {
-		if (strcmp(*wp, user) == 0)
+		if (strcmp(*wp, user) ||
+		    (icase && strcasecmp(*wp, user)) == 0)
 			return true;
 		++wp;
 	}
@@ -508,7 +498,7 @@ static const char *rc_debug(xmlNode *node, struct config *config,
 {
 	char *s;
 	if ((s = xmlGetProp_2s(node, "enable")) != NULL)
-		Debug = config->debug = strtol(s, NULL, 0);
+		Debug = config->debug = strtoul(s, NULL, 0);
 	free(s);
 	return NULL;
 }
@@ -551,7 +541,7 @@ static const char *rc_luserconf(xmlNode *node, struct config *config,
 	HX_strlcpy(config->luserconf, pent->pw_dir, sizeof(config->luserconf));
 	HX_strlcat(config->luserconf, "/", sizeof(config->luserconf));
 	HX_strlcat(config->luserconf, s, sizeof(config->luserconf));
-	w4rn("path to luserconf set to %s\n", config->luserconf),
+	w4rn("path to luserconf set to %s\n", config->luserconf);
 	free(s);
 	return NULL;
 }
@@ -569,7 +559,7 @@ static const char *rc_mkmountpoint(xmlNode *node, struct config *config,
 	return NULL;
 }
 
-/*
+/**
  * str_to_optlist -
  * @optlist:	destination list
  * @str:	string to parse
@@ -578,14 +568,49 @@ static const char *rc_mkmountpoint(xmlNode *node, struct config *config,
  * This is ok, since it is already an allocated string (i.e. does not
  * belong to libxml but to pam_mount). Caller frees it anyway right away.
  */
+static bool str_to_optkv(struct HXclist_head *optlist, char *str)
+{
+	char *value, *ptr;
+	struct kvp *kvp;
+
+	if (str == NULL || *str == '\0')
+		return true;
+
+	while ((ptr = HX_strsep(&str, ",")) != NULL) {
+		kvp = xmalloc(sizeof(struct kvp));
+		if (kvp == NULL)
+			return false;
+		HXlist_init(&kvp->list);
+		value = strchr(ptr, '=');
+		if (value != NULL) {
+			*value++ = '\0';
+			kvp->key   = xstrdup(ptr);
+			kvp->value = xstrdup(value);
+			if (kvp->key == NULL || kvp->value == NULL)
+				goto out;
+			HXclist_push(optlist, &kvp->list);
+		} else {
+			kvp->key = xstrdup(ptr);
+			if (kvp->key == NULL)
+				goto out;
+			kvp->value = NULL;
+			HXclist_push(optlist, &kvp->list);
+		}
+	}
+
+	return true;
+ out:
+	free(kvp->key);
+	free(kvp->value);
+	free(kvp);
+	return false;
+}
+
 static bool str_to_optlist(struct HXbtree *optlist, char *str)
 {
 	char *value, *ptr;
 
-	if (str == NULL || strlen(str) == 0)
-		/*
-		 * So what, ignore it.
-		 */
+	if (str == NULL || *str == '\0')
 		return true;
 
 	while ((ptr = HX_strsep(&str, ",")) != NULL) {
@@ -659,114 +684,509 @@ static const char *rc_string(xmlNode *node, struct config *config,
 	return NULL;
 }
 
-static const char *rc_volume_inter(struct config *config,
-    const struct volume_attrs *attr)
+/**
+ * rc_volume_cond_and - handle <and> element
+ * @pwd:	user logging in
+ * @node: 	XML <and> node
+ *
+ * Handle an <and> XML node, by processing all elements within it and ANDing
+ * them together. Returns negative on error.
+ */
+static int rc_volume_cond_and(const struct passwd *pwd, xmlNode *node)
+{
+	unsigned int count = 0;
+	int ret;
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_ELEMENT_NODE)
+			continue;
+		ret = rc_volume_cond_ext(pwd, node);
+		if (ret < 0)
+			return ret;
+		else if (ret == 0)
+			return false;
+		++count;
+	}
+
+	if (count > 0)
+		/*
+		 * If there was any non-matching element, we would have
+		 * already returned above.
+		 */
+		return true;
+
+	l0g("config: <and> does not have any child elements\n");
+	return -1;
+}
+
+/**
+ * rc_volume_cond_or - handle <or> element
+ * @pwd:	user logging in
+ * @node:	XML <or> node
+ *
+ * Handle an <or> XML node, by processing all elements within it and ORing
+ * them together. On error, returns -1.
+ */
+static int rc_volume_cond_or(const struct passwd *pwd, xmlNode *node)
+{
+	unsigned int count;
+	int ret;
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_ELEMENT_NODE)
+			continue;
+		ret = rc_volume_cond_ext(pwd, node);
+		if (ret < 0)
+			return ret;
+		else if (ret > 0)
+			return true;
+		++count;
+	}
+
+	if (count > 0)
+		return false;
+
+	l0g("config: <and> does not have any child elements\n");
+	return -1;
+}
+
+/**
+ * rc_volume_cond_xor - handle <xor> element
+ * @pwd:	user logging in
+ * @node:	XML <xor> node
+ *
+ * Handle an <xor> XML node, by processing the two elements within it and
+ * XORin them together. Exactly two elements must be provided, otherwise
+ * the function fails with -1.
+ */
+static int rc_volume_cond_xor(const struct passwd *pwd, xmlNode *node)
+{
+	unsigned int count = 0;
+	int ret[2];
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_ELEMENT_NODE)
+			continue;
+		if (count >= 2)
+			goto out;
+		ret[count] = rc_volume_cond_ext(pwd, node);
+		if (ret[count] < 0)
+			return ret[count];
+		++count;
+	}
+
+	if (count == 2)
+		return ret[0] ^ ret[1];
+ out:
+	l0g("config: <xor> must have exactly two child elements\n");
+	return -1;
+}
+
+/**
+ * rc_volume_cond_not - handle <not> element
+ * @pwd:	user logging in
+ * @node:	XML <not> node
+ *
+ * Handle an <not> XML node, by processing the single elements within it and
+ * negating it. Exactly one elements must be provided, otherwise the
+ * function fails with -1.
+ */
+static int rc_volume_cond_not(const struct passwd *pwd, xmlNode *node)
+{
+	unsigned int count = 0;
+	bool ret;
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_ELEMENT_NODE)
+			continue;
+		if (count++ > 0)
+			goto out;
+		ret = rc_volume_cond_ext(pwd, node);
+		if (ret < 0)
+			return ret;
+		ret = !ret;
+	}
+
+	if (count == 1)
+		return ret;
+ out:
+	l0g("config: <not> may only have one child element\n");
+	return -1;
+}
+
+/**
+ * rc_volume_cond_user - handle <user> element
+ * @pwd:	user logging in
+ * @node:	XML <user> node
+ */
+static int rc_volume_cond_user(const struct passwd *pwd, xmlNode *node)
+{
+	xmlNode *parent = node;
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_TEXT_NODE)
+			continue;
+		if (parse_bool_f(xmlGetProp_2s(parent, "icase")))
+			return strcasecmp(signed_cast(const char *,
+			       node->content), pwd->pw_name) == 0;
+		else
+			return strcmp_1u(node->content, pwd->pw_name) == 0;
+	}
+
+	return false;
+}
+
+static int __rc_volume_cond_id(const char *s, unsigned int id)
+{
+	unsigned int id_start, id_end;
+	char *delim;
+
+	id_start = id_end = strtoul(s, &delim, 0);
+	if (*delim == '\0')
+		return id_start == id;
+	if (*delim != '-' || *++delim == '\0')
+		return -1;
+	id_end = strtoul(delim, &delim, 0);
+	if (*delim != '\0')
+		return -1;
+	return id_start <= id && id <= id_end;
+}
+
+/**
+ * rc_volume_cond_uid - handle <uid> element
+ * @pwd:	user logging in
+ * @node:	XML <uid> node
+ */
+static int rc_volume_cond_uid(const struct passwd *pwd, xmlNode *node)
+{
+	int ret;
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_TEXT_NODE)
+			continue;
+		ret = __rc_volume_cond_id(signed_cast(const char *,
+		      node->content), pwd->pw_uid);
+		if (ret < 0)
+			return ret;
+	}
+
+	l0g("config: empty or invalid content for <%s>\n", "uid");
+	return -1;
+}
+
+/**
+ * rc_volume_cond_gid - handle <gid> element
+ * @pwd:	user logging in
+ * @node:	XML <uid> node
+ */
+static int rc_volume_cond_gid(const struct passwd *pwd, xmlNode *node)
+{
+	int ret;
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_TEXT_NODE)
+			continue;
+		ret = __rc_volume_cond_id(signed_cast(const char *,
+		      node->content), pwd->pw_gid);
+		if (ret < 0)
+			return ret;
+	}
+
+	l0g("config: empty or invalid content for <%s>\n", "gid");
+	return -1;
+}
+
+static int __rc_volume_cond_pgrp(const char *group, unsigned int gid,
+    bool icase)
+{
+	const struct group *grp;
+
+	errno = 0;
+	grp   = getgrgid(gid);
+	if (grp == NULL) {
+		if (errno == 0)
+			return 0; /* just not found */
+
+		w4rn("getgrgid(%u) failed: %s\n", gid, strerror(errno));
+		return -1;
+	}
+
+	if (icase)
+		return strcasecmp(group, grp->gr_name) == 0;
+	else
+		return strcmp(group, grp->gr_name) == 0;
+}
+
+/**
+ * rc_volume_cond_pgrp - handle <pgrp> element
+ * @pwd:	user logging in
+ * @node:	XML <pgrp> node (actually also <sgrp>)
+ */
+static int rc_volume_cond_pgrp(const struct passwd *pwd, xmlNode *node)
+{
+	xmlNode *parent = node;
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_TEXT_NODE)
+			continue;
+
+		return __rc_volume_cond_pgrp(
+		       signed_cast(const char *, node->content), pwd->pw_gid,
+		       parse_bool_f(xmlGetProp_2s(parent, "icase")));
+	}
+
+	l0g("config: empty or invalid content for <%s>\n", "pgrp");
+	return -1;
+}
+
+/**
+ * rc_volume_cond_sgrp - handle <sgrp> element
+ * @pwd:	user logging in
+ * @node:	XML <sgrp> node
+ */
+static int rc_volume_cond_sgrp(const struct passwd *pwd, xmlNode *node)
+{
+	const struct group *grp;
+	xmlNode *parent = node;
+	int ret;
+
+	for (node = node->children; node != NULL; node = node->next) {
+		if (node->type != XML_TEXT_NODE)
+			continue;
+
+		if ((grp = getgrgid(pwd->pw_gid)) == NULL) {
+			w4rn("getgrgid(%ld) failed: %s\n",
+			     static_cast(long, pwd->pw_gid), strerror(errno));
+			return -1;
+		}
+
+		ret = rc_volume_cond_pgrp(pwd, parent);
+		if (ret < 0 || ret > 0)
+			return ret;
+		return user_in_sgrp(pwd->pw_name,
+		       signed_cast(const char *, node->content),
+		       parse_bool_f(xmlGetProp_2s(parent, "icase")));
+	}
+
+	l0g("config: empty or invalid content for <%s>\n", "sgrp");
+	return -1;
+}
+
+/**
+ * rc_volume_cond_ext - interpret extended user control elements
+ * @pwd:	user logging in
+ * @node:	XML node to operate on
+ */
+static int rc_volume_cond_ext(const struct passwd *pwd, xmlNode *node)
+{
+	if (strcmp_1u(node->name, "and") == 0)
+		return rc_volume_cond_and(pwd, node);
+	else if (strcmp_1u(node->name, "or") == 0)
+		return rc_volume_cond_or(pwd, node);
+	else if (strcmp_1u(node->name, "xor") == 0)
+		return rc_volume_cond_xor(pwd, node);
+	else if (strcmp_1u(node->name, "not") == 0)
+		return rc_volume_cond_not(pwd, node);
+	else if (strcmp_1u(node->name, "user") == 0)
+		return rc_volume_cond_user(pwd, node);
+	else if (strcmp_1u(node->name, "uid") == 0)
+		return rc_volume_cond_uid(pwd, node);
+	else if (strcmp_1u(node->name, "gid") == 0)
+		return rc_volume_cond_gid(pwd, node);
+	else if (strcmp_1u(node->name, "pgrp") == 0)
+		return rc_volume_cond_pgrp(pwd, node);
+	else if (strcmp_1u(node->name, "sgrp") == 0)
+		return rc_volume_cond_sgrp(pwd, node);
+
+	l0g("config: unknown element <%s>\n", node->name);
+	return -1;
+}
+
+/**
+ * rc_volume_cond_simple - interpret simple user control elements
+ * @pwd:	user logging in
+ * @node:	XML <volume> node
+ */
+static int rc_volume_cond_simple(const struct passwd *pwd, xmlNode *node)
+{
+	char *user   = xmlGetProp_2s(node, "user");
+	char *invert = xmlGetProp_2s(node, "invert");
+	char *uid    = xmlGetProp_2s(node, "uid");
+	char *gid    = xmlGetProp_2s(node, "gid");
+	char *pgrp   = xmlGetProp_2s(node, "pgrp");
+	char *sgrp   = xmlGetProp_2s(node, "sgrp");
+	bool for_me  = true;
+	int ret      = true;
+
+	if (user == NULL && invert == NULL && uid == NULL && gid == NULL &&
+	    pgrp == NULL && sgrp == NULL)
+		return -1;
+
+	if (user != NULL) {
+		if (strcmp(user, "*") != 0)
+			for_me &= strcmp(pwd->pw_name, user) == 0;
+		else if (pwd->pw_uid == 0 || strcmp(pwd->pw_name, "root") == 0)
+			/* The wildcard never matches root */
+			for_me &= false;
+	}
+	if (uid != NULL) {
+		ret = __rc_volume_cond_id(uid, pwd->pw_uid);
+		if (ret < 0)
+			goto out;
+		for_me &= ret;
+	}
+	if (gid != NULL) {
+		ret = __rc_volume_cond_id(gid, pwd->pw_gid);
+		if (ret < 0)
+			goto out;
+		for_me &= ret;
+	}
+	if (pgrp != NULL) {
+		ret = __rc_volume_cond_pgrp(pgrp, pwd->pw_gid, false);
+		if (ret < 0)
+			goto out;
+		for_me &= ret;
+	}
+	if (sgrp != NULL) {
+		ret = __rc_volume_cond_pgrp(sgrp, pwd->pw_gid, false);
+		if (ret < 0)
+			goto out;
+		for_me &= ret;
+		ret = user_in_sgrp(pwd->pw_name, sgrp, false);
+		if (ret < 0)
+			goto out;
+		for_me &= ret;
+	}
+	if (invert != NULL) {
+		l0g("The \"invert\" attribute is deprecated, support will "
+		    "be removed in next version.\n");
+		if (strtoul(invert, NULL, 0))
+			for_me = !for_me;
+	}
+
+ out:
+	free(user);
+	free(invert);
+	free(uid);
+	free(gid);
+	free(pgrp);
+	free(sgrp);
+	if (ret < 0)
+		return ret;
+	return for_me;
+}
+
+/**
+ * rc_volume_cond - check if volume applies to user
+ * @node:	XML <volume> node
+ */
+static int rc_volume_cond(const char *user, xmlNode *node)
+{
+	struct passwd *pwd_ent;
+	int ret;
+
+	if ((pwd_ent = getpwnam(user)) == NULL) {
+		l0g("getpwnam: %s\n", strerror(errno));
+		return -1;
+	}
+
+	ret = rc_volume_cond_simple(pwd_ent, node);
+	if (ret < 0 && node->children != NULL) {
+		/* When no attributes, but elements... */
+		ret = rc_volume_cond_and(pwd_ent, node);
+		if (ret < 0)
+			return -1;
+		else if (ret == 0)
+			return false;
+	} else if (ret == 0) {
+		/* Attributes but (hopefully) no elements. */
+		if (node->children != NULL) {
+			l0g("You cannot have both simple and extended user control\n");
+			return -1;
+		}
+		return false;
+	}
+
+	return true;
+}
+
+static const char *rc_volume(xmlNode *node, struct config *config,
+    unsigned int command)
 {
-	enum wildcard_type wildcard = WC_NONE;
-	struct passwd *pent;
 	struct vol *vpt;
 	unsigned int i;
+	char *tmp;
 
-	if (strlen(attr->user) > sizeof_z(vpt->user) ||
-	    strlen(attr->fstype) > sizeof_z(vpt->fstype) ||
-	    strlen(attr->server) > sizeof_z(vpt->server) ||
-	    strlen(attr->path) > sizeof_z(vpt->volume) ||
-	    strlen(attr->mntpt) > sizeof_z(vpt->mountpoint) ||
-	    strlen(attr->fskeycipher) > sizeof_z(vpt->fs_key_cipher) ||
-	    strlen(attr->fskeypath) > sizeof_z(vpt->fs_key_path))
-		return "command too long";
-
-	if ((pent = getpwnam(config->user)) == NULL) {
-		w4rn("getpwnam(\"%s\") failed: %s\n",
-		     Config.user, strerror(errno));
+	if (strlen(config->user) > sizeof_z(vpt->user))
+		return "config: <volume> components too long\n";
+	if (rc_volume_cond(config->user, node) <= 0)
 		return NULL;
-	}
 
-	if (strcmp(attr->user, "*") == 0)
-		wildcard = WC_ANYUSER;
-	else if (*attr->pgrp != '\0')
-		wildcard = WC_PGRP;
-	else if (*attr->sgrp != '\0')
-		wildcard = WC_SGRP;
-	else if (*attr->user == '\0')
-		wildcard = WC_ANYUSER;
+	vpt = calloc(1, sizeof(struct vol));
+	if (vpt == NULL)
+		return strerror(errno);
 
-	if (wildcard != WC_NONE && (strcmp(config->user, "root") == 0 ||
-	    pent->pw_uid == 0)) {
-		/* One day, when SELinux becomes a daily thing, remove this. */
-		w4rn("volume wildcards ignored for \"root\" and uid0\n");
-		return NULL;
-	}
-
-	if (wildcard == WC_NONE && (strcmp(config->user, attr->user) != 0) ^
-	    attr->invert)
-		goto notforme;
-
-	if (wildcard == WC_PGRP) {
-		const char *grp_name = attr->pgrp;
-		struct group *gent;
-
-		if ((gent = getgrgid(pent->pw_gid)) == NULL) {
-			w4rn("getgrgid(%ld) failed: %s\n",
-			     static_cast(long, pent->pw_gid), strerror(errno));
-			return NULL;
-		}
-		if ((strcmp(grp_name, gent->gr_name) != 0) ^ attr->invert)
-			goto notforme;
-	} else if (wildcard == WC_SGRP) {
-		const char *grp_name = attr->sgrp;
-		struct group *gent;
-
-		if ((gent = getgrgid(pent->pw_gid)) == NULL) {
-			w4rn("getgrgid(%ld) failed: %s\n",
-			     static_cast(long, pent->pw_gid), strerror(errno));
-			return NULL;
-		}
-		if (!user_in_sgrp(config->user, grp_name) ^ attr->invert)
-			goto notforme;
-	}
-
-	/* realloc */
-	config->volume = xrealloc(config->volume,
-	                 sizeof(struct vol) * (config->volcount + 1));
-	vpt = &config->volume[config->volcount++];
-	memset(vpt, 0, sizeof(*vpt));
+	HXlist_init(&vpt->list);
+	HXclist_push(&config->volume_list, &vpt->list);
 
 	vpt->globalconf = config->level == CONTEXT_GLOBAL;
 	strncpy(vpt->user, config->user, sizeof(vpt->user));
 	vpt->type = CMD_LCLMOUNT;
-	vpt->options = HXbtree_init(HXBT_MAP | HXBT_CKEY | HXBT_CDATA |
-	               HXBT_SCMP | HXBT_CID);
+	HXclist_init(&vpt->options);
 
-	/* [1] */
-	strncpy(vpt->fstype, attr->fstype, sizeof(vpt->fstype));
-
-	for (i = 0; default_command[i].type != -1; ++i) {
-		const struct pmt_command *c = &default_command[i];
-		if (c->fs != NULL && strcasecmp(attr->fstype, c->fs) == 0) {
-			vpt->type = c->type;
-			break;
-		}
+	/* Eyeball ssh setting */
+	if ((tmp = xmlGetProp_2s(node, "ssh")) != NULL) {
+		vpt->uses_ssh = parse_bool(tmp);
+		free(tmp);
 	}
 
-	/* [2] */
-	if (*attr->server != '\0')
-		strncpy(vpt->server, attr->server, sizeof(vpt->server));
+	/* Filesystem type */
+	if ((tmp = xmlGetProp_2s(node, "fstype")) != NULL) {
+		if (strlen(tmp) > sizeof_z(vpt->fstype))
+			l0g("config: %s \"%s\" truncated\n", "fstype", tmp);
+		strncpy(vpt->fstype, tmp, sizeof(vpt->fstype));
 
-	/* [3] */
-	strncpy(vpt->volume, attr->path, sizeof(vpt->volume));
+		for (i = 0; default_command[i].type != -1; ++i) {
+			const struct pmt_command *c = &default_command[i];
+			if (c->fs != NULL && strcasecmp(tmp, c->fs) == 0) {
+				vpt->type = c->type;
+				break;
+			}
+		}
+		free(tmp);
+	} else {
+		strncpy(vpt->fstype, "auto", sizeof(vpt->fstype));
+	}
 
-	/* [4] */
-	if (*attr->mntpt == '\0') {
+	/* Source location */
+	if ((tmp = xmlGetProp_2s(node, "server")) != NULL) {
+		if (strlen(tmp) > sizeof_z(vpt->server))
+			l0g("config: %s \"%s\" truncated\n", "server", tmp);
+		strncpy(vpt->server, tmp, sizeof(vpt->server));
+		free(tmp);
+	}
+	if ((tmp = xmlGetProp_2s(node, "path")) != NULL) {
+		if (strlen(tmp) > sizeof_z(vpt->volume))
+			l0g("config: %s \"%s\" truncated\n", "path", tmp);
+		strncpy(vpt->volume, tmp, sizeof(vpt->volume));
+		free(tmp);
+	}
+
+	/* Destination */
+	if ((tmp = xmlGetProp_2s(node, "mountpoint")) != NULL) {
+		if (strlen(tmp) > sizeof_z(vpt->mountpoint))
+			l0g("config: %s \"%s\" truncated\n", "mountpoint", tmp);
+		strncpy(vpt->mountpoint, tmp, sizeof(vpt->mountpoint));
+		free(tmp);
+	} else {
 		if (!fstab_value(vpt->volume, FSTAB_MNTPT, vpt->mountpoint,
 		    sizeof(vpt->mountpoint)))
-				return "could not determine mountpoint";
+			return "could not determine mountpoint";
 		vpt->use_fstab = 1;
-	} else {
-		strncpy(vpt->mountpoint, attr->mntpt, sizeof(vpt->mountpoint));
 	}
 
-	if (*attr->options == '\0') {
+	/* Options */
+	if ((tmp = xmlGetProp_2s(node, "options")) == NULL) {
 		/*
 		 * Three options: field defined, field is '-' and fstab should
 		 * be used (when no mount point was provided either) or field
@@ -777,73 +1197,33 @@ static const char *rc_volume_inter(struct config *config,
 			if (!fstab_value(vpt->volume, FSTAB_OPTS, options,
 			    sizeof(options)))
 				return "could not determine options";
-			if (!str_to_optlist(vpt->options, options))
+			if (!str_to_optkv(&vpt->options, options))
 				return "error parsing mount options";
 		}
-	} else if (!str_to_optlist(vpt->options, attr->options)) {
+	} else if (!str_to_optkv(&vpt->options, tmp)) {
+		free(tmp);
 		return "error parsing mount options";
+	} else {
+		free(tmp);
 	}
 
-	if (*attr->fskeycipher != '\0')
-		strncpy(vpt->fs_key_cipher, attr->fskeycipher,
-		        sizeof(vpt->fs_key_cipher));
-	if (*attr->fskeypath != '\0')
-		strncpy(vpt->fs_key_path, attr->fskeypath,
-		        sizeof(vpt->fs_key_path));
+	/* Filesystem key */
+	if ((tmp = xmlGetProp_2s(node, "fskeycipher")) != NULL) {
+		if (strlen(tmp) > sizeof_z(vpt->fs_key_cipher))
+			l0g("config: %s \"%s\" truncated\n", "fskeycipher", tmp);
+		strncpy(vpt->fs_key_cipher, tmp, sizeof(vpt->fs_key_cipher));
+		free(tmp);
+	}
+	if ((tmp = xmlGetProp_2s(node, "fskeypath")) != NULL) {
+		if (strlen(tmp) > sizeof_z(vpt->fs_key_path))
+			l0g("config: %s \"%s\" truncated\n", "fskeypath", tmp);
+		strncpy(vpt->fs_key_path, tmp, sizeof(vpt->fs_key_path));
+		free(tmp);
+	}
 
 	/* expandconfig() will set this later */
-	vpt->used_wildcard = 0;
+	vpt->used_wildcard = false;
 	return NULL;
-
- notforme:
-	w4rn("ignoring volume record... (not for me)\n");
-	return NULL;
-}
-
-static const char *rc_volume(xmlNode *node, struct config *config,
-    unsigned int command)
-{
-	struct volume_attrs norm, orig = {
-		.user        = xmlGetProp_2s(node, "user"),
-		.pgrp        = xmlGetProp_2s(node, "pgrp"),
-		.sgrp        = xmlGetProp_2s(node, "sgrp"),
-		.fstype      = xmlGetProp_2s(node, "fstype"),
-		.server      = xmlGetProp_2s(node, "server"),
-		.path        = xmlGetProp_2s(node, "path"),
-		.mntpt       = xmlGetProp_2s(node, "mountpoint"),
-		.options     = xmlGetProp_2s(node, "options"),
-		.fskeycipher = xmlGetProp_2s(node, "fskeycipher"),
-		.fskeypath   = xmlGetProp_2s(node, "fskeypath"),
-	};
-	const char *ret;
-	char *invert;
-
-	if ((invert = xmlGetProp_2s(node, "invert")) != NULL) {
-		orig.invert = strtoul(invert, NULL, 0);
-		free(invert);
-	}
-
-	memcpy(&norm, &orig, sizeof(norm));
-	if (norm.user        == NULL) norm.user        = "";
-	if (norm.pgrp        == NULL) norm.pgrp        = "";
-	if (norm.sgrp        == NULL) norm.sgrp        = "";
-	if (norm.fstype      == NULL) norm.fstype      = "auto";
-	if (norm.server      == NULL) norm.server      = "";
-	if (norm.path        == NULL) norm.path        = "";
-	if (norm.mntpt       == NULL) norm.mntpt       = "";
-	if (norm.options     == NULL) norm.options     = "";
-	if (norm.fskeycipher == NULL) norm.fskeycipher = "";
-	if (norm.fskeypath   == NULL) norm.fskeypath   = "";
-	ret = rc_volume_inter(config, &norm);
-	free(orig.user);
-	free(orig.fstype);
-	free(orig.server);
-	free(orig.path);
-	free(orig.mntpt);
-	free(orig.options);
-	free(orig.fskeycipher);
-	free(orig.fskeypath);
-	return ret;
 }
 
 //-----------------------------------------------------------------------------
@@ -860,7 +1240,7 @@ static const struct pmt_command default_command[] = {
 	{CMD_NFSMOUNT,   "nfs",   "nfsmount",   {"mount", "%(SERVER):%(VOLUME)", "%(MNTPT)%(before=\"-o\" OPTIONS)", NULL}},
 	{CMD_LCLMOUNT,   NULL,    "lclmount",   {"mount", "-p0", "-t", "%(FSTYPE)", "%(VOLUME)", "%(MNTPT)", "%(before=\"-o\" OPTIONS)", NULL}},
 	/* Hope to have this in util-linux (LCLMOUNT) some day: */
-	{CMD_CRYPTMOUNT, "crypt", "cryptmount", {"mount", "-t", "crypt", "%(before=\"-o \" OPTIONS)", "%(VOLUME)", "%(MNTPT)", NULL}},
+	{CMD_CRYPTMOUNT, "crypt", "cryptmount", {"mount.crypt", "%(before=\"-o \" OPTIONS)", "%(VOLUME)", "%(MNTPT)", NULL}},
 	{CMD_UMOUNT,     NULL,    "umount",     {"umount", "%(MNTPT)", NULL}},
 	{CMD_LSOF,       NULL,    "lsof",       {"lsof", "%(MNTPT)", NULL}},
 	/*
@@ -884,6 +1264,7 @@ static const struct callbackmap cf_tags[] = {
 	{"cryptmount",      rc_command,             CMD_CRYPTMOUNT},
 	{"davmount",        rc_command,             CMD_DAVMOUNT},
 	{"debug",           rc_debug,               CMD_NONE},
+	{"fd0ssh",          rc_command,             CMD_FD0SSH},
 	{"fsckloop",        rc_fsckloop,            CMD_NONE},
 	{"fsck",            rc_command,             CMD_FSCK},
 	{"fusemount",       rc_command,             CMD_FUSEMOUNT},
@@ -911,5 +1292,3 @@ static const struct callbackmap cf_tags[] = {
 	{"volume",          rc_volume,              CMD_NONE},
 	{NULL},
 };
-
-//=============================================================================
